@@ -27,10 +27,11 @@ from mypy.types import (
 )
 from mypy.typetraverser import TypeTraverserVisitor
 from mypy.nodes import (
-    TypeInfo, Context, MypyFile, op_methods, FuncDef, reverse_builtin_aliases,
+    TypeInfo, Context, MypyFile, op_methods, op_methods_to_symbols,
+    FuncDef, reverse_builtin_aliases,
     ARG_POS, ARG_OPT, ARG_NAMED, ARG_NAMED_OPT, ARG_STAR, ARG_STAR2,
     ReturnStmt, NameExpr, Var, CONTRAVARIANT, COVARIANT, SymbolNode,
-    CallExpr
+    CallExpr, SymbolTable
 )
 from mypy.subtypes import (
     is_subtype, find_member, get_member_flags,
@@ -174,7 +175,12 @@ class MessageBuilder:
     # get some information as arguments, and they build an error message based
     # on them.
 
-    def has_no_attr(self, original_type: Type, typ: Type, member: str, context: Context) -> Type:
+    def has_no_attr(self,
+                    original_type: Type,
+                    typ: Type,
+                    member: str,
+                    context: Context,
+                    module_symbol_table: Optional[SymbolTable] = None) -> Type:
         """Report a missing or non-accessible member.
 
         original_type is the top-level type on which the error occurred.
@@ -183,6 +189,11 @@ class MessageBuilder:
         will be the specific item in the union that does not have the member
         attribute.
 
+        'module_symbol_table' is passed to this function if the type for which we
+        are trying to get a member was originally a module. The SymbolTable allows
+        us to look up and suggests attributes of the module since they are not
+        directly available on original_type
+
         If member corresponds to an operator, use the corresponding operator
         name in the messages. Return type Any.
         """
@@ -225,7 +236,7 @@ class MessageBuilder:
             self.fail('Unsupported target for indexed assignment', context, code=codes.INDEX)
         elif member == '__call__':
             if isinstance(original_type, Instance) and \
-                    (original_type.type.fullname() == 'builtins.function'):
+                    (original_type.type.fullname == 'builtins.function'):
                 # "'function' not callable" is a confusing error message.
                 # Explain that the problem is that the type of the function is not known.
                 self.fail('Cannot call function of unknown type', context, code=codes.OPERATOR)
@@ -243,6 +254,15 @@ class MessageBuilder:
                 failed = False
                 if isinstance(original_type, Instance) and original_type.type.names:
                     alternatives = set(original_type.type.names.keys())
+
+                    if module_symbol_table is not None:
+                        alternatives |= {key for key in module_symbol_table.keys()}
+
+                    # in some situations, the member is in the alternatives set
+                    # but since we're in this function, we shouldn't suggest it
+                    if member in alternatives:
+                        alternatives.remove(member)
+
                     matches = [m for m in COMMON_MISTAKES.get(member, []) if m in alternatives]
                     matches.extend(best_matches(member, alternatives)[:3])
                     if member == '__aiter__' and matches == ['__iter__']:
@@ -353,7 +373,7 @@ class MessageBuilder:
             else:
                 base = extract_type(name)
 
-            for op, method in op_methods.items():
+            for method, op in op_methods_to_symbols.items():
                 for variant in method, '__r' + method[2:]:
                     # FIX: do not rely on textual formatting
                     if name.startswith('"{}" of'.format(variant)):
@@ -366,6 +386,11 @@ class MessageBuilder:
                                                            context, code=codes.OPERATOR)
                         return codes.OPERATOR
 
+            if name.startswith('"__cmp__" of'):
+                self.unsupported_operand_types("comparison", arg_type, base,
+                                               context, code=codes.OPERATOR)
+                return codes.INDEX
+
             if name.startswith('"__getitem__" of'):
                 self.invalid_index_type(arg_type, callee.arg_types[n - 1], base, context,
                                         code=codes.INDEX)
@@ -488,7 +513,7 @@ class MessageBuilder:
                 expected_types = list(expected_type.items)
             else:
                 expected_types = [expected_type]
-            for type in expected_types:
+            for type in get_proper_types(expected_types):
                 if isinstance(arg_type, Instance) and isinstance(type, Instance):
                     notes = append_invariance_notes(notes, arg_type, type)
         self.fail(msg, context, code=code)
@@ -507,7 +532,8 @@ class MessageBuilder:
             self.report_protocol_problems(original_caller_type, callee_type, context, code=code)
         if (isinstance(callee_type, CallableType) and
                 isinstance(original_caller_type, Instance)):
-            call = find_member('__call__', original_caller_type, original_caller_type)
+            call = find_member('__call__', original_caller_type, original_caller_type,
+                               is_operator=True)
             if call:
                 self.note_call(original_caller_type, call, context, code=code)
 
@@ -522,7 +548,7 @@ class MessageBuilder:
         if (argument_names is not None and not all(k is None for k in argument_names)
                 and len(argument_names) >= 1):
             num_positional_args = sum(k is None for k in argument_names)
-            arguments_left = callee.arg_names[num_positional_args:]
+            arguments_left = callee.arg_names[num_positional_args:callee.min_args]
             diff = [k for k in arguments_left if k not in argument_names]
             if len(diff) == 1:
                 msg = 'Missing positional argument'
@@ -563,9 +589,24 @@ class MessageBuilder:
         msg = 'Too many positional arguments' + for_function(callee)
         self.fail(msg, context)
 
-    def unexpected_keyword_argument(self, callee: CallableType, name: str,
+    def unexpected_keyword_argument(self, callee: CallableType, name: str, arg_type: Type,
                                     context: Context) -> None:
         msg = 'Unexpected keyword argument "{}"'.format(name) + for_function(callee)
+        # Suggest intended keyword, look for type match else fallback on any match.
+        matching_type_args = []
+        not_matching_type_args = []
+        for i, kwarg_type in enumerate(callee.arg_types):
+            callee_arg_name = callee.arg_names[i]
+            if callee_arg_name is not None and callee.arg_kinds[i] != ARG_STAR:
+                if is_subtype(arg_type, kwarg_type):
+                    matching_type_args.append(callee_arg_name)
+                else:
+                    not_matching_type_args.append(callee_arg_name)
+        matches = best_matches(name, matching_type_args)
+        if not matches:
+            matches = best_matches(name, not_matching_type_args)
+        if matches:
+            msg += "; did you mean {}?".format(pretty_or(matches[:3]))
         self.fail(msg, context, code=codes.CALL_ARG)
         module = find_defining_module(self.modules, callee)
         if module:
@@ -574,7 +615,7 @@ class MessageBuilder:
             if not fname:  # an alias to function with a different name
                 fname = 'Called function'
             self.note('{} defined here'.format(fname), callee.definition,
-                      file=module.path, origin=context)
+                      file=module.path, origin=context, code=codes.CALL_ARG)
 
     def duplicate_argument_value(self, callee: CallableType, index: int,
                                  context: Context) -> None:
@@ -778,25 +819,33 @@ class MessageBuilder:
         self.fail('Argument 1 for "super" must be a type object; got {}'.format(type_str), context)
 
     def too_few_string_formatting_arguments(self, context: Context) -> None:
-        self.fail('Not enough arguments for format string', context)
+        self.fail('Not enough arguments for format string', context,
+                  code=codes.STRING_FORMATTING)
 
     def too_many_string_formatting_arguments(self, context: Context) -> None:
-        self.fail('Not all arguments converted during string formatting', context)
+        self.fail('Not all arguments converted during string formatting', context,
+                  code=codes.STRING_FORMATTING)
 
     def unsupported_placeholder(self, placeholder: str, context: Context) -> None:
-        self.fail('Unsupported format character \'%s\'' % placeholder, context)
+        self.fail('Unsupported format character \'%s\'' % placeholder, context,
+                  code=codes.STRING_FORMATTING)
 
     def string_interpolation_with_star_and_key(self, context: Context) -> None:
-        self.fail('String interpolation contains both stars and mapping keys', context)
+        self.fail('String interpolation contains both stars and mapping keys', context,
+                  code=codes.STRING_FORMATTING)
 
-    def requires_int_or_char(self, context: Context) -> None:
-        self.fail('%c requires int or char', context)
+    def requires_int_or_char(self, context: Context,
+                             format_call: bool = False) -> None:
+        self.fail('"{}c" requires int or char'.format(':' if format_call else '%'),
+                  context, code=codes.STRING_FORMATTING)
 
     def key_not_in_mapping(self, key: str, context: Context) -> None:
-        self.fail('Key \'%s\' not found in mapping' % key, context)
+        self.fail('Key \'%s\' not found in mapping' % key, context,
+                  code=codes.STRING_FORMATTING)
 
     def string_interpolation_mixing_key_and_non_keys(self, context: Context) -> None:
-        self.fail('String interpolation mixes specifier with and without mapping keys', context)
+        self.fail('String interpolation mixes specifier with and without mapping keys', context,
+                  code=codes.STRING_FORMATTING)
 
     def cannot_determine_type(self, name: str, context: Context) -> None:
         self.fail("Cannot determine type of '%s'" % name, context, code=codes.HAS_TYPE)
@@ -832,7 +881,7 @@ class MessageBuilder:
                                             context: Context) -> None:
         self.fail('Definition of "{}" in base class "{}" is incompatible '
                   'with definition in base class "{}"'.format(
-                      name, base1.name(), base2.name()), context)
+                      name, base1.name, base2.name), context)
 
     def cant_assign_to_method(self, context: Context) -> None:
         self.fail(message_registry.CANNOT_ASSIGN_TO_METHOD, context,
@@ -865,7 +914,7 @@ class MessageBuilder:
     def read_only_property(self, name: str, type: TypeInfo,
                            context: Context) -> None:
         self.fail('Property "{}" defined in "{}" is read-only'.format(
-            name, type.name()), context)
+            name, type.name), context)
 
     def incompatible_typevar_value(self,
                                    callee: CallableType,
@@ -917,17 +966,18 @@ class MessageBuilder:
                   'of signature {}'.format(index), context)
 
     def warn_both_operands_are_from_unions(self, context: Context) -> None:
-        self.note('Both left and right operands are unions', context)
+        self.note('Both left and right operands are unions', context, code=codes.OPERATOR)
 
     def warn_operand_was_from_union(self, side: str, original: Type, context: Context) -> None:
-        self.note('{} operand is of type {}'.format(side, format_type(original)), context)
+        self.note('{} operand is of type {}'.format(side, format_type(original)), context,
+                  code=codes.OPERATOR)
 
     def operator_method_signatures_overlap(
             self, reverse_class: TypeInfo, reverse_method: str, forward_class: Type,
             forward_method: str, context: Context) -> None:
         self.fail('Signatures of "{}" of "{}" and "{}" of {} '
                   'are unsafely overlapping'.format(
-                      reverse_method, reverse_class.name(),
+                      reverse_method, reverse_class.name,
                       forward_method, format_type(forward_class)),
                   context)
 
@@ -981,23 +1031,23 @@ class MessageBuilder:
         has_variable_annotations = not python_version or python_version >= (3, 6)
         # Only gives hint if it's a variable declaration and the partial type is a builtin type
         if (python_version and isinstance(node, Var) and isinstance(node.type, PartialType) and
-                node.type.type and node.type.type.fullname() in reverse_builtin_aliases):
-            alias = reverse_builtin_aliases[node.type.type.fullname()]
+                node.type.type and node.type.type.fullname in reverse_builtin_aliases):
+            alias = reverse_builtin_aliases[node.type.type.fullname]
             alias = alias.split('.')[-1]
             type_dec = '<type>'
             if alias == 'Dict':
                 type_dec = '{}, {}'.format(type_dec, type_dec)
             if has_variable_annotations:
-                hint = ' (hint: "{}: {}[{}] = ...")'.format(node.name(), alias, type_dec)
+                hint = ' (hint: "{}: {}[{}] = ...")'.format(node.name, alias, type_dec)
             else:
-                hint = ' (hint: "{} = ...  # type: {}[{}]")'.format(node.name(), alias, type_dec)
+                hint = ' (hint: "{} = ...  # type: {}[{}]")'.format(node.name, alias, type_dec)
 
         if has_variable_annotations:
             needed = 'annotation'
         else:
             needed = 'comment'
 
-        self.fail("Need type {} for '{}'{}".format(needed, unmangle(node.name()), hint), context,
+        self.fail("Need type {} for '{}'{}".format(needed, unmangle(node.name), hint), context,
                   code=codes.VAR_ANNOTATED)
 
     def explicit_any(self, ctx: Context) -> None:
@@ -1094,6 +1144,18 @@ class MessageBuilder:
             format_type(typ))
         self.fail(message, context, code=codes.NO_ANY_RETURN)
 
+    def incorrect__exit__return(self, context: Context) -> None:
+        self.fail(
+            '"bool" is invalid as return type for "__exit__" that always returns False', context,
+            code=codes.EXIT_RETURN)
+        self.note(
+            'Use "typing_extensions.Literal[False]" as the return type or change it to "None"',
+            context, code=codes.EXIT_RETURN)
+        self.note(
+            'If return type of "__exit__" implies that it may return True, '
+            'the context manager may swallow exceptions',
+            context, code=codes.EXIT_RETURN)
+
     def untyped_decorated_function(self, typ: Type, context: Context) -> None:
         typ = get_proper_type(typ)
         if isinstance(typ, AnyType):
@@ -1123,7 +1185,7 @@ class MessageBuilder:
 
     def cannot_use_function_with_type(
             self, method_name: str, type_name: str, context: Context) -> None:
-        self.fail("Cannot use {}() with a {} type".format(method_name, type_name), context)
+        self.fail("Cannot use {}() with {} type".format(method_name, type_name), context)
 
     def report_non_method_protocol(self, tp: TypeInfo, members: List[str],
                                    context: Context) -> None:
@@ -1132,7 +1194,7 @@ class MessageBuilder:
         if len(members) < 3:
             attrs = ', '.join(members)
             self.note('Protocol "{}" has non-method member(s): {}'
-                      .format(tp.name(), attrs), context)
+                      .format(tp.name, attrs), context)
 
     def note_call(self,
                   subtype: Type,
@@ -1193,7 +1255,7 @@ class MessageBuilder:
         exclusions = {TypedDictType: ['typing.Mapping'],
                       TupleType: ['typing.Iterable', 'typing.Sequence'],
                       Instance: []}  # type: Dict[type, List[str]]
-        if supertype.type.fullname() in exclusions[type(subtype)]:
+        if supertype.type.fullname in exclusions[type(subtype)]:
             return
         if any(isinstance(tp, UninhabitedType) for tp in get_proper_types(supertype.args)):
             # We don't want to add notes for failed inference (e.g. Iterable[<nothing>]).
@@ -1214,7 +1276,7 @@ class MessageBuilder:
         if (missing and len(missing) < len(supertype.type.protocol_members) and
                 len(missing) <= MAX_ITEMS):
             self.note("'{}' is missing following '{}' protocol member{}:"
-                      .format(subtype.type.name(), supertype.type.name(), plural_s(missing)),
+                      .format(subtype.type.name, supertype.type.name, plural_s(missing)),
                       context,
                       code=code)
             self.note(', '.join(missing), context, offset=OFFSET, code=code)
@@ -1261,22 +1323,22 @@ class MessageBuilder:
         for name, subflags, superflags in conflict_flags[:MAX_ITEMS]:
             if IS_CLASSVAR in subflags and IS_CLASSVAR not in superflags:
                 self.note('Protocol member {}.{} expected instance variable,'
-                          ' got class variable'.format(supertype.type.name(), name),
+                          ' got class variable'.format(supertype.type.name, name),
                           context,
                           code=code)
             if IS_CLASSVAR in superflags and IS_CLASSVAR not in subflags:
                 self.note('Protocol member {}.{} expected class variable,'
-                          ' got instance variable'.format(supertype.type.name(), name),
+                          ' got instance variable'.format(supertype.type.name, name),
                           context,
                           code=code)
             if IS_SETTABLE in superflags and IS_SETTABLE not in subflags:
                 self.note('Protocol member {}.{} expected settable variable,'
-                          ' got read-only attribute'.format(supertype.type.name(), name),
+                          ' got read-only attribute'.format(supertype.type.name, name),
                           context,
                           code=code)
             if IS_CLASS_OR_STATIC in superflags and IS_CLASS_OR_STATIC not in subflags:
                 self.note('Protocol member {}.{} expected class or static method'
-                          .format(supertype.type.name(), name),
+                          .format(supertype.type.name, name),
                           context,
                           code=code)
         self.print_more(conflict_flags, context, OFFSET, MAX_ITEMS, code=code)
@@ -1375,42 +1437,37 @@ def format_type_inner(typ: Type,
     if isinstance(typ, Instance):
         itype = typ
         # Get the short name of the type.
-        if itype.type.fullname() in ('types.ModuleType',
-                                     '_importlib_modulespec.ModuleType'):
+        if itype.type.fullname in ('types.ModuleType', '_importlib_modulespec.ModuleType'):
             # Make some common error messages simpler and tidier.
             return 'Module'
-        if verbosity >= 2 or (fullnames and itype.type.fullname() in fullnames):
-            base_str = itype.type.fullname()
+        if verbosity >= 2 or (fullnames and itype.type.fullname in fullnames):
+            base_str = itype.type.fullname
         else:
-            base_str = itype.type.name()
+            base_str = itype.type.name
         if itype.args == []:
             # No type arguments, just return the type name
             return base_str
-        elif itype.type.fullname() == 'builtins.tuple':
+        elif itype.type.fullname == 'builtins.tuple':
             item_type_str = format(itype.args[0])
             return 'Tuple[{}, ...]'.format(item_type_str)
-        elif itype.type.fullname() in reverse_builtin_aliases:
-            alias = reverse_builtin_aliases[itype.type.fullname()]
+        elif itype.type.fullname in reverse_builtin_aliases:
+            alias = reverse_builtin_aliases[itype.type.fullname]
             alias = alias.split('.')[-1]
             items = [format(arg) for arg in itype.args]
             return '{}[{}]'.format(alias, ', '.join(items))
         else:
             # There are type arguments. Convert the arguments to strings.
-            # If the result is too long, replace arguments with [...].
             a = []  # type: List[str]
             for arg in itype.args:
                 a.append(format(arg))
             s = ', '.join(a)
-            if len((base_str + s)) < 150:
-                return '{}[{}]'.format(base_str, s)
-            else:
-                return '{}[...]'.format(base_str)
+            return '{}[{}]'.format(base_str, s)
     elif isinstance(typ, TypeVarType):
         # This is similar to non-generic instance types.
         return typ.name
     elif isinstance(typ, TupleType):
         # Prefer the name of the fallback class (if not tuple), as it's more informative.
-        if typ.partial_fallback.type.fullname() != 'builtins.tuple':
+        if typ.partial_fallback.type.fullname != 'builtins.tuple':
             return format(typ.partial_fallback)
         items = []
         for t in typ.items:
@@ -1441,9 +1498,10 @@ def format_type_inner(typ: Type,
     elif isinstance(typ, UnionType):
         # Only print Unions as Optionals if the Optional wouldn't have to contain another Union
         print_as_optional = (len(typ.items) -
-                             sum(isinstance(t, NoneType) for t in typ.items) == 1)
+                             sum(isinstance(get_proper_type(t), NoneType)
+                                 for t in typ.items) == 1)
         if print_as_optional:
-            rest = [t for t in typ.items if not isinstance(t, NoneType)]
+            rest = [t for t in typ.items if not isinstance(get_proper_type(t), NoneType)]
             return 'Optional[{}]'.format(format(rest[0]))
         else:
             items = []
@@ -1540,7 +1598,7 @@ def find_type_overlaps(*types: Type) -> Set[str]:
     d = {}  # type: Dict[str, Set[str]]
     for type in types:
         for inst in collect_all_instances(type):
-            d.setdefault(inst.type.name(), set()).add(inst.type.fullname())
+            d.setdefault(inst.type.name, set()).add(inst.type.fullname)
 
     overlaps = set()  # type: Set[str]
     for fullnames in d.values():
@@ -1630,14 +1688,14 @@ def pretty_callable(tp: CallableType) -> str:
             s += ' = ...'
 
     # If we got a "special arg" (i.e: self, cls, etc...), prepend it to the arg list
-    if isinstance(tp.definition, FuncDef) and tp.definition.name() is not None:
+    if isinstance(tp.definition, FuncDef) and tp.definition.name is not None:
         definition_args = tp.definition.arg_names
         if definition_args and tp.arg_names != definition_args \
                 and len(definition_args) > 0:
             if s:
                 s = ', ' + s
             s = definition_args[0] + s
-        s = '{}({})'.format(tp.definition.name(), s)
+        s = '{}({})'.format(tp.definition.name, s)
     elif tp.name:
         first_arg = tp.def_extras.get('first_arg')
         if first_arg:
@@ -1654,7 +1712,7 @@ def pretty_callable(tp: CallableType) -> str:
         for tvar in tp.variables:
             upper_bound = get_proper_type(tvar.upper_bound)
             if (isinstance(upper_bound, Instance) and
-                    upper_bound.type.fullname() != 'builtins.object'):
+                    upper_bound.type.fullname != 'builtins.object'):
                 tvars.append('{} <: {}'.format(tvar.name, format_type_bare(upper_bound)))
             elif tvar.values:
                 tvars.append('{} in ({})'
@@ -1800,7 +1858,7 @@ def for_function(callee: CallableType) -> str:
 def find_defining_module(modules: Dict[str, MypyFile], typ: CallableType) -> Optional[MypyFile]:
     if not typ.definition:
         return None
-    fullname = typ.definition.fullname()
+    fullname = typ.definition.fullname
     if fullname is not None and '.' in fullname:
         for i in range(fullname.count('.')):
             module_name = fullname.rsplit('.', i + 1)[0]
@@ -1843,13 +1901,13 @@ def append_invariance_notes(notes: List[str], arg_type: Instance,
     """Explain that the type is invariant and give notes for how to solve the issue."""
     invariant_type = ''
     covariant_suggestion = ''
-    if (arg_type.type.fullname() == 'builtins.list' and
-            expected_type.type.fullname() == 'builtins.list' and
+    if (arg_type.type.fullname == 'builtins.list' and
+            expected_type.type.fullname == 'builtins.list' and
             is_subtype(arg_type.args[0], expected_type.args[0])):
         invariant_type = 'List'
         covariant_suggestion = 'Consider using "Sequence" instead, which is covariant'
-    elif (arg_type.type.fullname() == 'builtins.dict' and
-          expected_type.type.fullname() == 'builtins.dict' and
+    elif (arg_type.type.fullname == 'builtins.dict' and
+          expected_type.type.fullname == 'builtins.dict' and
           is_same_type(arg_type.args[0], expected_type.args[0]) and
           is_subtype(arg_type.args[1], expected_type.args[1])):
         invariant_type = 'Dict'
@@ -1863,8 +1921,10 @@ def append_invariance_notes(notes: List[str], arg_type: Instance,
     return notes
 
 
-def make_inferred_type_note(context: Context, subtype: Type,
-                            supertype: Type, supertype_str: str) -> str:
+def make_inferred_type_note(context: Context,
+                            subtype: Type,
+                            supertype: Type,
+                            supertype_str: str) -> str:
     """Explain that the user may have forgotten to type a variable.
 
     The user does not expect an error if the inferred container type is the same as the return
@@ -1876,7 +1936,7 @@ def make_inferred_type_note(context: Context, subtype: Type,
     supertype = get_proper_type(supertype)
     if (isinstance(subtype, Instance) and
             isinstance(supertype, Instance) and
-            subtype.type.fullname() == supertype.type.fullname() and
+            subtype.type.fullname == supertype.type.fullname and
             subtype.args and
             supertype.args and
             isinstance(context, ReturnStmt) and
