@@ -1,30 +1,37 @@
-import glob
+import io
 import os.path
 import shutil
 import sys
 import tempfile
 import re
+import unittest
 from types import ModuleType
 
-from typing import List, Tuple
+from typing import Any, List, Tuple, Optional
 
-from mypy.test.helpers import Suite, assert_equal, assert_string_arrays_equal
+from mypy.test.helpers import (
+    assert_equal, assert_string_arrays_equal, local_sys_path_set
+)
 from mypy.test.data import DataSuite, DataDrivenTestCase
 from mypy.errors import CompileError
 from mypy.stubgen import (
     generate_stubs, parse_options, Options, collect_build_targets,
-    mypy_options
+    mypy_options, is_blacklisted_path, is_non_library_module
 )
-from mypy.stubutil import walk_packages
+from mypy.stubutil import walk_packages, remove_misplaced_type_comments, common_dir_prefix
 from mypy.stubgenc import generate_c_type_stub, infer_method_sig, generate_c_function_stub
 from mypy.stubdoc import (
     parse_signature, parse_all_signatures, build_signature, find_unique_signatures,
     infer_sig_from_docstring, infer_prop_type_from_docstring, FunctionSig, ArgSig,
-    infer_arg_sig_from_docstring
+    infer_arg_sig_from_docstring, is_valid_type
 )
+from mypy.moduleinspect import ModuleInspect, InspectError
 
 
-class StubgenCmdLineSuite(Suite):
+class StubgenCmdLineSuite(unittest.TestCase):
+    """Test cases for processing command-line options and finding files."""
+
+    @unittest.skipIf(sys.platform == 'win32', "clean up fails on Windows")
     def test_files_found(self) -> None:
         current = os.getcwd()
         with tempfile.TemporaryDirectory() as tmp:
@@ -45,6 +52,7 @@ class StubgenCmdLineSuite(Suite):
             finally:
                 os.chdir(current)
 
+    @unittest.skipIf(sys.platform == 'win32', "clean up fails on Windows")
     def test_packages_found(self) -> None:
         current = os.getcwd()
         with tempfile.TemporaryDirectory() as tmp:
@@ -64,33 +72,56 @@ class StubgenCmdLineSuite(Suite):
             finally:
                 os.chdir(current)
 
+    @unittest.skipIf(sys.platform == 'win32', "clean up fails on Windows")
+    def test_module_not_found(self) -> None:
+        current = os.getcwd()
+        captured_output = io.StringIO()
+        sys.stdout = captured_output
+        with tempfile.TemporaryDirectory() as tmp:
+            try:
+                os.chdir(tmp)
+                self.make_file(tmp, 'mymodule.py', content='import a')
+                opts = parse_options(['-m', 'mymodule'])
+                py_mods, c_mods = collect_build_targets(opts, mypy_options(opts))
+                assert captured_output.getvalue() == ''
+            finally:
+                sys.stdout = sys.__stdout__
+                os.chdir(current)
+
     def make_file(self, *path: str, content: str = '') -> None:
         file = os.path.join(*path)
         with open(file, 'w') as f:
             f.write(content)
 
+    def run(self, result: Optional[Any] = None) -> Optional[Any]:
+        with local_sys_path_set():
+            return super().run(result)
 
-class StubgenCliParseSuite(Suite):
+
+class StubgenCliParseSuite(unittest.TestCase):
     def test_walk_packages(self) -> None:
-        assert_equal(
-            set(walk_packages(["mypy.errors"])),
-            {"mypy.errors"})
+        with ModuleInspect() as m:
+            assert_equal(
+                set(walk_packages(m, ["mypy.errors"])),
+                {"mypy.errors"})
 
-        assert_equal(
-            set(walk_packages(["mypy.errors", "mypy.stubgen"])),
-            {"mypy.errors", "mypy.stubgen"})
+            assert_equal(
+                set(walk_packages(m, ["mypy.errors", "mypy.stubgen"])),
+                {"mypy.errors", "mypy.stubgen"})
 
-        all_mypy_packages = set(walk_packages(["mypy"]))
-        self.assertTrue(all_mypy_packages.issuperset({
-            "mypy",
-            "mypy.errors",
-            "mypy.stubgen",
-            "mypy.test",
-            "mypy.test.helpers",
-        }))
+            all_mypy_packages = set(walk_packages(m, ["mypy"]))
+            self.assertTrue(all_mypy_packages.issuperset({
+                "mypy",
+                "mypy.errors",
+                "mypy.stubgen",
+                "mypy.test",
+                "mypy.test.helpers",
+            }))
 
 
-class StubgenUtilSuite(Suite):
+class StubgenUtilSuite(unittest.TestCase):
+    """Unit tests for stubgen utility functions."""
+
     def test_parse_signature(self) -> None:
         self.assert_parse_signature('func()', ('func', [], []))
 
@@ -232,6 +263,13 @@ class StubgenUtilSuite(Suite):
                      [FunctionSig(name='func', args=[ArgSig(name='x'), ArgSig(name='y')],
                                   ret_type='int')])
 
+    def test_infer_sig_from_docstring_bad_indentation(self) -> None:
+        assert_equal(infer_sig_from_docstring("""
+            x
+              x
+             x
+            """, 'func'), None)
+
     def test_infer_arg_sig_from_docstring(self) -> None:
         assert_equal(infer_arg_sig_from_docstring("(*args, **kwargs)"),
                      [ArgSig(name='*args'), ArgSig(name='**kwargs')])
@@ -249,24 +287,264 @@ class StubgenUtilSuite(Suite):
                      'Tuple[int, int]')
         assert_equal(infer_prop_type_from_docstring('\nstr: A string.'), None)
 
+    def test_infer_sig_from_docstring_square_brackets(self) -> None:
+        assert infer_sig_from_docstring(
+            'fetch_row([maxrows, how]) -- Fetches stuff',
+            'fetch_row',
+        ) == []
+
+    def test_remove_misplaced_type_comments_1(self) -> None:
+        good = """
+        \u1234
+        def f(x):  # type: (int) -> int
+
+        def g(x):
+            # type: (int) -> int
+
+        def h():
+
+            # type: () int
+
+        x = 1  # type: int
+        """
+
+        assert_equal(remove_misplaced_type_comments(good), good)
+
+    def test_remove_misplaced_type_comments_2(self) -> None:
+        bad = """
+        def f(x):
+            # type: Callable[[int], int]
+            pass
+
+        #  type:  "foo"
+        #  type:  'bar'
+        x = 1
+        # type: int
+        """
+        bad_fixed = """
+        def f(x):
+
+            pass
+
+
+
+        x = 1
+
+        """
+        assert_equal(remove_misplaced_type_comments(bad), bad_fixed)
+
+    def test_remove_misplaced_type_comments_3(self) -> None:
+        bad = '''
+        def f(x):
+            """docstring"""
+            # type: (int) -> int
+            pass
+
+        def g(x):
+            """docstring
+            """
+            # type: (int) -> int
+            pass
+        '''
+        bad_fixed = '''
+        def f(x):
+            """docstring"""
+
+            pass
+
+        def g(x):
+            """docstring
+            """
+
+            pass
+        '''
+        assert_equal(remove_misplaced_type_comments(bad), bad_fixed)
+
+    def test_remove_misplaced_type_comments_4(self) -> None:
+        bad = """
+        def f(x):
+            '''docstring'''
+            # type: (int) -> int
+            pass
+
+        def g(x):
+            '''docstring
+            '''
+            # type: (int) -> int
+            pass
+        """
+        bad_fixed = """
+        def f(x):
+            '''docstring'''
+
+            pass
+
+        def g(x):
+            '''docstring
+            '''
+
+            pass
+        """
+        assert_equal(remove_misplaced_type_comments(bad), bad_fixed)
+
+    def test_remove_misplaced_type_comments_5(self) -> None:
+        bad = """
+        def f(x):
+            # type: (int, List[Any],
+            #        float, bool) -> int
+            pass
+
+        def g(x):
+            # type: (int, List[Any])
+            pass
+        """
+        bad_fixed = """
+        def f(x):
+
+            #        float, bool) -> int
+            pass
+
+        def g(x):
+
+            pass
+        """
+        assert_equal(remove_misplaced_type_comments(bad), bad_fixed)
+
+    def test_remove_misplaced_type_comments_bytes(self) -> None:
+        original = b"""
+        \xbf
+        def f(x):  # type: (int) -> int
+
+        def g(x):
+            # type: (int) -> int
+            pass
+
+        def h():
+            # type: int
+            pass
+
+        x = 1  # type: int
+        """
+
+        dest = b"""
+        \xbf
+        def f(x):  # type: (int) -> int
+
+        def g(x):
+            # type: (int) -> int
+            pass
+
+        def h():
+
+            pass
+
+        x = 1  # type: int
+        """
+
+        assert_equal(remove_misplaced_type_comments(original), dest)
+
+    def test_common_dir_prefix(self) -> None:
+        assert common_dir_prefix([]) == '.'
+        assert common_dir_prefix(['x.pyi']) == '.'
+        assert common_dir_prefix(['./x.pyi']) == '.'
+        assert common_dir_prefix(['foo/bar/x.pyi']) == 'foo/bar'
+        assert common_dir_prefix(['foo/bar/x.pyi',
+                                  'foo/bar/y.pyi']) == 'foo/bar'
+        assert common_dir_prefix(['foo/bar/x.pyi', 'foo/y.pyi']) == 'foo'
+        assert common_dir_prefix(['foo/x.pyi', 'foo/bar/y.pyi']) == 'foo'
+        assert common_dir_prefix(['foo/bar/zar/x.pyi', 'foo/y.pyi']) == 'foo'
+        assert common_dir_prefix(['foo/x.pyi', 'foo/bar/zar/y.pyi']) == 'foo'
+        assert common_dir_prefix(['foo/bar/zar/x.pyi', 'foo/bar/y.pyi']) == 'foo/bar'
+        assert common_dir_prefix(['foo/bar/x.pyi', 'foo/bar/zar/y.pyi']) == 'foo/bar'
+
+
+class StubgenHelpersSuite(unittest.TestCase):
+    def test_is_blacklisted_path(self) -> None:
+        assert not is_blacklisted_path('foo/bar.py')
+        assert not is_blacklisted_path('foo.py')
+        assert not is_blacklisted_path('foo/xvendor/bar.py')
+        assert not is_blacklisted_path('foo/vendorx/bar.py')
+        assert is_blacklisted_path('foo/vendor/bar.py')
+        assert is_blacklisted_path('foo/vendored/bar.py')
+        assert is_blacklisted_path('foo/vendored/bar/thing.py')
+        assert is_blacklisted_path('foo/six.py')
+
+    def test_is_non_library_module(self) -> None:
+        assert not is_non_library_module('foo')
+        assert not is_non_library_module('foo.bar')
+
+        # The following could be test modules, but we are very conservative and
+        # don't treat them as such since they could plausibly be real modules.
+        assert not is_non_library_module('foo.bartest')
+        assert not is_non_library_module('foo.bartests')
+        assert not is_non_library_module('foo.testbar')
+
+        assert is_non_library_module('foo.test')
+        assert is_non_library_module('foo.test.foo')
+        assert is_non_library_module('foo.tests')
+        assert is_non_library_module('foo.tests.foo')
+        assert is_non_library_module('foo.testing.foo')
+        assert is_non_library_module('foo.SelfTest.foo')
+
+        assert is_non_library_module('foo.test_bar')
+        assert is_non_library_module('foo.bar_tests')
+        assert is_non_library_module('foo.testing')
+        assert is_non_library_module('foo.conftest')
+        assert is_non_library_module('foo.bar_test_util')
+        assert is_non_library_module('foo.bar_test_utils')
+        assert is_non_library_module('foo.bar_test_base')
+
+        assert is_non_library_module('foo.setup')
+
+        assert is_non_library_module('foo.__main__')
+
 
 class StubgenPythonSuite(DataSuite):
+    """Data-driven end-to-end test cases that generate stub files.
+
+    You can use these magic test case name suffixes:
+
+    *_semanal
+        Run semantic analysis (slow as this uses real stubs -- only use
+        when necessary)
+    *_import
+        Import module and perform runtime introspection (in the current
+        process!)
+
+    You can use these magic comments:
+
+    # flags: --some-stubgen-option ...
+        Specify custom stubgen options
+
+    # modules: module1 module2 ...
+        Specify which modules to output (by default only 'main')
+    """
+
     required_out_section = True
     base_path = '.'
     files = ['stubgen.test']
 
     def run_case(self, testcase: DataDrivenTestCase) -> None:
-        extra = []
-        mods = []
+        with local_sys_path_set():
+            self.run_case_inner(testcase)
+
+    def run_case_inner(self, testcase: DataDrivenTestCase) -> None:
+        extra = []  # Extra command-line args
+        mods = []  # Module names to process
         source = '\n'.join(testcase.input)
         for file, content in testcase.files + [('./main.py', source)]:
-            mod = os.path.basename(file)[:-3]
+            # Strip ./ prefix and .py suffix.
+            mod = file[2:-3].replace('/', '.')
+            if mod.endswith('.__init__'):
+                mod, _, _ = mod.rpartition('.')
             mods.append(mod)
-            extra.extend(['-m', mod])
+            if '-p ' not in source:
+                extra.extend(['-m', mod])
             with open(file, 'w') as f:
                 f.write(content)
 
         options = self.parse_flags(source, extra)
+        modules = self.parse_modules(source)
         out_dir = 'out'
         try:
             try:
@@ -274,9 +552,11 @@ class StubgenPythonSuite(DataSuite):
                     options.no_import = True
                 if not testcase.name.endswith('_semanal'):
                     options.parse_only = True
-                generate_stubs(options, quiet=True, add_header=False)
+                generate_stubs(options)
                 a = []  # type: List[str]
-                self.add_file(os.path.join(out_dir, 'main.pyi'), a)
+                for module in modules:
+                    fnam = module_to_path(out_dir, module)
+                    self.add_file(fnam, a, header=len(modules) > 1)
             except CompileError as e:
                 a = e.messages
             assert_string_arrays_equal(testcase.output, a,
@@ -294,32 +574,57 @@ class StubgenPythonSuite(DataSuite):
             flag_list = flags.group(1).split()
         else:
             flag_list = []
-        return parse_options(flag_list + extra)
+        options = parse_options(flag_list + extra)
+        if '--verbose' not in flag_list:
+            options.quiet = True
+        else:
+            options.verbose = True
+        return options
 
-    def add_file(self, path: str, result: List[str]) -> None:
+    def parse_modules(self, program_text: str) -> List[str]:
+        modules = re.search('# modules: (.*)$', program_text, flags=re.MULTILINE)
+        if modules:
+            return modules.group(1).split()
+        else:
+            return ['main']
+
+    def add_file(self, path: str, result: List[str], header: bool) -> None:
+        if not os.path.exists(path):
+            result.append('<%s was not generated>' % path.replace('\\', '/'))
+            return
+        if header:
+            result.append('# {}'.format(path[4:]))
         with open(path, encoding='utf8') as file:
             result.extend(file.read().splitlines())
 
 
-class StubgencSuite(Suite):
+self_arg = ArgSig(name='self')
+
+
+class StubgencSuite(unittest.TestCase):
+    """Unit tests for stub generation from C modules using introspection.
+
+    Note that these don't cover a lot!
+    """
+
     def test_infer_hash_sig(self) -> None:
-        assert_equal(infer_method_sig('__hash__'), [])
+        assert_equal(infer_method_sig('__hash__'), [self_arg])
 
     def test_infer_getitem_sig(self) -> None:
-        assert_equal(infer_method_sig('__getitem__'), [ArgSig(name='index')])
+        assert_equal(infer_method_sig('__getitem__'), [self_arg, ArgSig(name='index')])
 
     def test_infer_setitem_sig(self) -> None:
         assert_equal(infer_method_sig('__setitem__'),
-                     [ArgSig(name='index'), ArgSig(name='object')])
+                     [self_arg, ArgSig(name='index'), ArgSig(name='object')])
 
     def test_infer_binary_op_sig(self) -> None:
         for op in ('eq', 'ne', 'lt', 'le', 'gt', 'ge',
                    'add', 'radd', 'sub', 'rsub', 'mul', 'rmul'):
-            assert_equal(infer_method_sig('__%s__' % op), [ArgSig(name='other')])
+            assert_equal(infer_method_sig('__%s__' % op), [self_arg, ArgSig(name='other')])
 
     def test_infer_unary_op_sig(self) -> None:
         for op in ('neg', 'pos'):
-            assert_equal(infer_method_sig('__%s__' % op), [])
+            assert_equal(infer_method_sig('__%s__' % op), [self_arg])
 
     def test_generate_c_type_stub_no_crash_for_object(self) -> None:
         output = []  # type: List[str]
@@ -500,7 +805,7 @@ class StubgencSuite(Suite):
         assert_equal(set(imports), {'from typing import overload'})
 
 
-class ArgSigSuite(Suite):
+class ArgSigSuite(unittest.TestCase):
     def test_repr(self) -> None:
         assert_equal(repr(ArgSig(name='asd"dsa')),
                      "ArgSig(name='asd\"dsa', type=None, default=False)")
@@ -510,3 +815,68 @@ class ArgSigSuite(Suite):
                      "ArgSig(name='func', type='str', default=False)")
         assert_equal(repr(ArgSig("func", 'str', default=True)),
                      "ArgSig(name='func', type='str', default=True)")
+
+
+class IsValidTypeSuite(unittest.TestCase):
+    def test_is_valid_type(self) -> None:
+        assert is_valid_type('int')
+        assert is_valid_type('str')
+        assert is_valid_type('Foo_Bar234')
+        assert is_valid_type('foo.bar')
+        assert is_valid_type('List[int]')
+        assert is_valid_type('Dict[str, int]')
+        assert is_valid_type('None')
+        assert not is_valid_type('foo-bar')
+        assert not is_valid_type('x->y')
+        assert not is_valid_type('True')
+        assert not is_valid_type('False')
+        assert not is_valid_type('x,y')
+        assert not is_valid_type('x, y')
+
+
+class ModuleInspectSuite(unittest.TestCase):
+    def test_python_module(self) -> None:
+        with ModuleInspect() as m:
+            p = m.get_package_properties('inspect')
+            assert p is not None
+            assert p.name == 'inspect'
+            assert p.file
+            assert p.path is None
+            assert p.is_c_module is False
+            assert p.subpackages == []
+
+    def test_python_package(self) -> None:
+        with ModuleInspect() as m:
+            p = m.get_package_properties('unittest')
+            assert p is not None
+            assert p.name == 'unittest'
+            assert p.file
+            assert p.path
+            assert p.is_c_module is False
+            assert p.subpackages
+            assert all(sub.startswith('unittest.') for sub in p.subpackages)
+
+    def test_c_module(self) -> None:
+        with ModuleInspect() as m:
+            p = m.get_package_properties('_socket')
+            assert p is not None
+            assert p.name == '_socket'
+            assert p.file
+            assert p.path is None
+            assert p.is_c_module is True
+            assert p.subpackages == []
+
+    def test_non_existent(self) -> None:
+        with ModuleInspect() as m:
+            with self.assertRaises(InspectError) as e:
+                m.get_package_properties('foobar-non-existent')
+            assert str(e.exception) == "No module named 'foobar-non-existent'"
+
+
+def module_to_path(out_dir: str, module: str) -> str:
+    fnam = os.path.join(out_dir, '{}.pyi'.format(module.replace('.', '/')))
+    if not os.path.exists(fnam):
+        alt_fnam = fnam.replace('.pyi', '/__init__.pyi')
+        if os.path.exists(alt_fnam):
+            return alt_fnam
+    return fnam
